@@ -18,24 +18,23 @@ use base qw( Exporter );
 our %EXPORT_TAGS = (
     'ALL'     => [ qw( array_to_moose
                        throw_nonunique_keys throw_multiple_rows
-                       set_class_ind set_key_ind                 ) ],
-    'TESTING' => [ qw( _check_descriptor _check_subobj
-                       _check_ref_attribs _check_non_ref_attribs ) ],
+                       set_class_ind set_key_ind               ) ],
+    'TESTING' => [ qw( _check_attribs _check_types             ) ],
 );
 
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'ALL'} }, @{ $EXPORT_TAGS{'TESTING'} } );
 
-our @EXPORT = qw( array_to_moose 
-
+our @EXPORT = qw(
+	
 );
 
-use version; our $VERSION = qv('0.0.9');
+our $VERSION = '0.01';
 
 # BEGIN { $Exporter::Verbose=1 };
 
 #BEGIN { print "Got Array::To:Moose Module\n" }
 
-use Params::Validate::Array qw(:all);
+use Params::Validate qw(:all);
 use Array::GroupBy qw(igroup_by str_row_equal);
 use Carp;
 use Data::Dumper;
@@ -89,187 +88,46 @@ sub throw_nonunique_keys { $throw_nonunique_keys = defined $_[0] ? $_[0] : 1 }
 ########################################
 sub throw_multiple_rows { $throw_multiple_rows = defined $_[0] ? $_[0] : 1 }
 
-##########
-# Usage
-#   my $moose_object_ref = array_to_moose( data => $array_ref,
-#                                          desc => { ... },
-#                                        );
 ############################################
-sub array_to_moose {
-  my ($data, $desc) = validate(@_,
-                          [ data => { type => ARRAYREF },
-                            desc => { type => HASHREF  },
-                          ]
-  );
-
-  croak "'data => ...' isn't a 2D array (AoA)"
-    unless ref($data->[0]);
-
-  croak 'empty descriptor'
-    unless keys %$desc;
-
-  #print "data ", Dumper($data), "\ndesc ", Dumper($desc);
-
-
-  my $result = [];   # returned result is either an array or a hash of objects
-
-  # extract column of possible hash key
-  my $keycol;
-
-  if (exists $desc->{$KEY}) {
-    $keycol = $desc->{$KEY};
-
-    $result = {};         # returning a hashref
-
-  }
-
-  # _check_descriptor returns:
-  # $class,       the class of the object
-  # $attribs,     a hashref (attrib => column_number) of "simple" attributes
-  #               (column numbers only)
-  # $ref_attribs, a hashref of attribute/column number values for
-  #               non-simple attributes, currently limited to "ArrayRef[`a]",
-  #               where `a is e.g 'Str', etc (i.e. `a is not a class)
-  # $sub_desc,    a hashref of sub-objects.
-  #               the keys are the attrib. names, the values the
-  #               descriptors of the next level down
-
-  my ($class, $attribs, $ref_attribs, $sub_obj_desc) =
-            _check_descriptor($data, $desc);
-
-  #print "data ", Dumper($data), "\nattrib = ", Dumper($attribs),
-  #      "\nargs = ", Dumper([ values %$attribs ]);
-
-  #print "\$ref_attribs ", Dumper($ref_attribs); exit;
-
-  my $iter = igroup_by(
-                data    => $data,
-                compare => \&str_row_equal,
-                args    => [ values %$attribs ],
-  );
-
-  while (my $subset = $iter->()) {
-
-    #print "subset: ", Dumper($subset), "\n";
-
-    #print "before 1: attrib ", Dumper($attribs), "\ndata ", Dumper($subset);
-
-    # change attribs from col numbers to values:
-    # from:  { name => 1,           sex => 2,      ... }
-    # to     { name => 'Smith, J.', sex => 'male', ... }
-    my %attribs = map { $_ => $subset->[0]->[$attribs->{$_}] } keys %$attribs;
-   
-
-    # print "after 1: attrib ", Dumper(\%attribs), "\n";
-
-    # add the 'simple ArrayRef' sub-objects
-    # (there should really be only one of these - test for it?)
-    while (my($attr_name, $col) = each %$ref_attribs) {
-      my @col = map { $_->[$col] } @$subset;
-      $attribs{$attr_name} = \@col;
-
-      # ... or ...
-      #$attribs{$attr_name} = [ map { $_->[$col] } @$subset ];
-    }
-
-    # print "after 2: attrib ", Dumper(\%attribs), "\n";
-
-    # sub-objects - recursive call to array_to_moose()
-    while( my($attr_name, $desc) = each %$sub_obj_desc) {
-
-      my $type = $class->meta->find_attribute_by_name($attr_name)->type_constraint
-        or croak "Moose attribute '$attr_name' has no type";
-
-      #print "'$attr_name' has type '$type'";
-
-      my $sub_obj = array_to_moose( data => $subset,
-                                    desc => $desc,
-                                  );
-
-      $sub_obj = _check_subobj($class, $attr_name, $type, $sub_obj);
-
-      #print "type $type\n";
-
-      $attribs{$attr_name} = $sub_obj;
-    }
-
-    # print "after 2: attrib ", Dumper(\%attribs), "\n";
-
-    my $obj;
-    eval { $obj = $class->meta->new_object(%attribs) };
-    croak "Can't make a new '$class' object:\n$@\n"
-          if $@;
-
-    if (defined $keycol) {
-      my $key_name = $subset->[0]->[$keycol];
-
-      # optionally croak if we are overwriting an existing hash entry
-      croak "Non-unique key '$key_name' in '", $desc->{$CLASS}, "' class"
-        if exists $result->{$key_name} and $throw_nonunique_keys;
-
-      $result->{$key_name} = $obj;
-    } else {
-      push @{$result}, $obj;
-    }
-  }
-  return $result;
-}
-
+# Usage: my ($meta, $attrib, $sub_obj_desc)
+#                  = _check_attribs($data, $desc)
+#
+# Redo###
+# Throws an exception if the class has not been defined, or if the
+# attributes whose names are the keys of %my_attribs are not in the class
+# Also throws exception if 'class' (or 'key' if used in the
+# "keyword => " data) is an attribute of the class.
+#
+# Returns the class meta object
 ############################################
-# Usage: my ($class, $attribs, $ref_attribs, $sub_desc)
-#                  = _check_descriptor($data, $desc)
-#
-# Check the correctness of the descriptor hashref, $desc.
-#
-# Checks of descriptor $desc include:
-# 1. "class => 'MyClass'" line exists, and that class "MyClass" has
-#                         been defined
-# 2. for "attrib => N" 
-#     or "key    => N" lines, N, the column number, is an integer, and that
-#                      the column numbers is within limits of the data
-# 3. For "attrib => [N]", (note square brackets), N, the columnn number,
-#                         is within limits of the data
-#
-# Returns:
-# $class,      the class name,
-# $attribs,    hashref (name => column_index) of "simple" attributes
-# $ref_attribs hashref (name => column_index) of attribs which are
-#               ArrayRef[']s of simple types (i.e. not a Class)
-#               (HashRef[']s not implemented)
-# $sub_desc    hashref (name => desc) of sub-object descriptors
-############################################
-sub _check_descriptor {
+sub _check_attribs {
   my ($data, $desc) = @_;
 
-  # remove from production!
-  croak "_check_descriptor() needs two arguments"
+  croak "_check_attribs() needs two arguments"
     unless @_ == 2;
 
   my $class = $desc->{$CLASS}
-    or croak "No class descriptor '$CLASS => ...' in descriptor:\n",
-       Dumper($desc);
+    or croak "No class descriptor '$CLASS => ...' in descriptor:\n", Dumper($desc);
 
   my $meta;
 
-  # see other example of getting meta in Moose::Manual::???
   eval{ $meta = $class->meta };
-  croak "Class '$class' not defined: $@"
+  croak "Class '$class' not defined"
     if $@;
 
   my $ncols = @{ $data->[0] };
 
-  # separate out simple (i.e. non-reference) attributes, reference
-  # attributes, and sub-objects
-  my ($attrib, $ref_attrib, $sub_desc);
+  # separate out the attributes and optional sub-objects
+  my ($attrib, $sub_obj_desc);
 
   while ( my ($name, $value) =  each %$desc) {
 
-    # check lines which have 'simple' column numbers ( attrib or key => N)
+    # check column ranges
     unless (ref($value) or $name eq $CLASS) {
 
       my $msg = "attribute '$name => $value'";
 
-      croak "$msg must be a (non-negative) integer"
+      croak "$msg must be a +ve integer"
         unless $value =~ /^\d+$/;
 
       croak "$msg greater than # cols in the data ($ncols)"
@@ -280,91 +138,151 @@ sub _check_descriptor {
     if ($name eq $CLASS or $name eq $KEY) {
       croak "The '$class' object has an attribute called '$name'"
         if $meta->find_attribute_by_name($name);
-
       next;
     }
 
     croak "Attribute '$name' not in '$class' object"
       unless $meta->find_attribute_by_name($name);
 
-    if ((my $ref = ref($value)) eq 'HASH') {
-      $sub_desc->{$name} = $value;
+    my $ref = ref($value);
 
-    } elsif ($ref eq 'ARRAY') {
-      # descr entry looks like, e.g.:
-      #   attrib => [6],
-      #
-      # ( or attrib => [key => 6, value => 7],  in future... ?)
-
-      croak "attribute must be of form, e.g.: '$name => [N], "
-            . "where N is a single integer'"
-          unless @$value == 1;
-
-      my $msg = "attribute '$name => [ " . $value->[0] . " ]'. '" .
-                  $value->[0] . "'";
-
-      croak "$msg must be a (non-negative) integer"
-        unless $value->[0]  =~ /^\d+$/;
-
-      croak "$msg greater than # cols in the data ($ncols)"
-        if $value->[0] > $ncols - 1;
-
-      $ref_attrib->{$name} = $value->[0];
+    if ($ref eq 'HASH') {
+      $sub_obj_desc->{$name} = $value;
 
     } elsif ($ref) {
       croak "attribute '$name' can't be a '$ref' reference";
 
     } else {
-      # "simple" attribute
       $attrib->{$name} = $value;
     }
   }
 
-
-  # check ref- and ...
-  _check_ref_attribs($class, $ref_attrib)
-    if $ref_attrib;
-
-  # ... non-ref attributes from the descriptor against the Moose object
-  _check_non_ref_attribs($class, $attrib)
-    if $attrib;
-
   croak "no attributes with column numbers in descriptor:\n", Dumper($desc)
     unless $attrib and %$attrib;
 
-  return ($class, $attrib, $ref_attrib, $sub_desc);
+  return ($meta, $attrib, $sub_obj_desc);
+}
+
+##########  
+# Usage
+#   my $moose_object_ref = array_to_moose( data => $array_ref,
+#                                          desc => { ... }, 
+#                                        );
+#
+# This is the new version which figgers out what the attributes should be
+# from the type of the Moose attributes, e.g., 'Arrayref[...]', 'HashRef[...]',
+# etc
+############################################
+sub array_to_moose {
+  # to do this in one step is *ugly*: (...) = @{ { v() } }{ qw(...) }
+  # so leave as two steps
+  my %opts = validate(@_, { data => { type => ARRAYREF },
+                            desc => { type => HASHREF  },
+                          }
+  );
+
+  my ($data, $desc) = @opts{ qw(data desc) };
+
+  croak "'data => ...' isn't a 2D array (AoA)"
+    unless ref($data->[0]);
+
+  croak 'empty descriptor'
+    unless keys %$desc;
+
+  #print "data ", Dumper($data), "\ndesc ", Dumper($desc);
+
+
+  my $result = [];        # returned result either an array or a hash of objects
+
+  # extract column of possible hash key
+  my $keycol;
+
+  if (exists $desc->{$KEY}) {
+    $keycol = $desc->{$KEY};
+
+    $result = {};         # we will be returning a hashref
+
+  }
+
+  # $attrib is a hash of attribute/column number values 
+  # $sub_obj_desc is a (possible empty) hash of sub-objects descriptors
+  # the keys are the attrib. names, the value is the hash ref of the
+  # descriptor of the next level down
+
+  my ($meta, $attrib, $sub_obj_desc) = _check_attribs($data, $desc);
+
+  # print "data ", Dumper($data), "\nattrib = ", Dumper($attrib),
+  #       "\nargs = ", Dumper([ values %$attrib ]);
+
+  my $iter = igroup_by(
+                data    => $data,
+                compare => \&str_row_equal,
+                args    => [ values %$attrib ],
+  );
+
+  while (my $subset = $iter->()) {
+
+    #print "subset: ", Dumper($subset), "\n";
+
+    # hash of Moose sub-objects: key is the attribute name, value the Moose
+    # sub-object
+    my %sub_objs;  # hash of Moose sub-objects
+
+    # make the sub-objects first
+    while( my($attr_name, $desc) = each %$sub_obj_desc) {
+
+      my $type = $meta->get_attribute($attr_name)->type_constraint
+        or croak "Moose attribute '$attr_name' has no type";
+
+      #print "'$attr_name' has type '$type'";
+
+      my $sub_obj = array_to_moose( data => $subset,
+                                    desc => $desc,
+                                  );
+
+      $sub_obj = _check_types($attr_name, $type, $sub_obj);
+
+      #print "type $type\n";
+
+      $sub_objs{$attr_name} = $sub_obj;
+    }
+
+    #print "before: attrib ", Dumper($attrib), "\ndata ", Dumper($subset);
+
+    # change %attrib from: { name => 1, sex => 2, ...}
+    # to { name => 'Smith, J.', sex => 'male', ... }
+    my %attrib = map { $_ => $subset->[0]->[$attrib->{$_}] } keys %$attrib;
+
+    # print "after attrib ", Dumper(\%attrib), "\n";
+
+    my $obj;
+    eval { $obj = $meta->new_object(%attrib, %sub_objs) };
+    croak "Can't make a new '" . $desc->{$CLASS} . "' object:\n$@\n"
+          if $@;
+
+    if (defined $keycol) {
+      my $key_name = $subset->[0]->[$keycol];
+
+      # optionally croak if we are overwriting an existing hash entry
+      croak "Non-unique key '$key_name' in '", $desc->{$CLASS}, "' class"
+        if exists $result->{$key_name} and $throw_nonunique_keys;
+        
+      $result->{$key_name} = $obj;
+    } else {  
+      push @{$result}, $obj;
+    }
+  }
+  return $result;
 }
 
 ########################################
-# Usage: $sub_obj = _check_subobj($class, $attr_name, $type, $sub_obj);
+# Usage: $sub_obj = _check_types($attr_name, $type, $sub_obj);
 #
-# $class        is the name of the current class
-# $attr_name    is the name of the attribute in the descriptor, e.g.
-#               MyObjs => { ... } (used only diagnostic messages)
-# $type         is the expected Moose type of the sub-object
-#               i.e. 'HashRef[MyObj]', 'ArrayRef[MyObj]', or 'MyObj'
-# $sub_obj_ref  Reference to the data (just returned from a recursive call to
-#               array_to_moose() ) to be stored in the sub-object,
-#               i.e. isa => 'HashRef[MyObj]', isa => 'ArrayRef[MyObj]',
-#               or isa => 'MyObj'
-#
-#
-# Checks that the data in $sub_obj_ref agrees with the type of the object to
-# contain it
-# if $type is a ref to an object (isa => 'MyObj'), _check_subobj() converts
-# $sub_obj_ref from an arrayref to sub-object to ref to a subobj
-# (see notes in code below)
-#
-# Throws error is it finds a type mis-match
+# Checks that the descriptor's idea of the object type ($type) agrees
+# with the Moose declaration.
 ########################################
-sub _check_subobj {
-  my ($class, $attr_name, $type, $sub_obj) = @_;
-
-  # for now...
-  croak "_check_subobj() should have 4 args" unless @_ == 4;
-
-  #my $type = $class->meta->find_attribute_by_name($attr_name)->type_constraint
-  #  or croak "Moose class '$class' attribute '$attr_name' has no type";
+sub _check_types {
+  my ($attr_name, $type, $sub_obj) = @_;
 
   if ( $type =~ /^HashRef\[([^]]*)\]/ ) {
 
@@ -391,22 +309,17 @@ sub _check_subobj {
           . "of type '" . ref($sub_obj) . "'\n"
       if ref($sub_obj) ne 'ARRAY';
 
-    croak "Moose attribute '$attr_name' has type '$type' "
-          . "but your descriptor produced an object "
+    croak "Moose attribute '$attr_name' has type '$type'"
+          . "but your descriptor produced an object"
           . "of type 'ArrayRef[" . ref( $sub_obj->[0] ) . "]'\n"
       if ref( $sub_obj->[0] ) ne $1;
 
   } else {
 
-    # not isa => 'ArrayRef[MyObj]' or 'HashRef[MyObj]' but isa => 'MyObj',
-    # *but* since array_to_moose() can return only a hash- or arrayref of Moose
-    # objects, $sub_obj will be an arrayref of Moose objects, which we convert to a
-    # ref to an object
+    # not an ArrayRef[] or HashRef[] but an Obj ref
 
-    croak "Moose attribute '$attr_name' has type '$type' "
-          . "but your descriptor generated a '"
-          . ref($sub_obj)
-          . "' object and not the expected ARRAY"
+    croak "desc generated a '" . ref $sub_obj 
+          . "' object and not the expected array"
       unless ref $sub_obj eq 'ARRAY';
 
     # optionally give error if we got more than one row
@@ -414,144 +327,18 @@ sub _check_subobj {
         scalar @$sub_obj, " of them"
       if @$sub_obj != 1 and $throw_multiple_rows;
 
-    # convert from arrayref of objects to ref to object
     $sub_obj = $sub_obj->[0];
 
     # print "\$sub_obj type is ", ref($sub_obj), "\n";
 
-    croak "Moose attribute '$attr_name' has type '$type' "
-          . "but your descriptor produced an object "
-          . "of type '" . ref( $sub_obj ) . "'"
+    croak "Moose attribute '$attr_name' has type '$type'\n"
+          . "but your descriptor produced an object\n"
+          . "of type '" . ref( $sub_obj ) . "'\n"
       unless ref( $sub_obj ) eq $type;
   }
   return $sub_obj;
 }
 
-{
-
-  # The Moose type hierarchy (from Moose::Manual::Types) is:
-  # Any
-  # Item
-  #     Bool
-  #     Maybe[`a]
-  #     Undef
-  #     Defined
-  #         Value
-  #             Str
-  #                 Num
-  #                     Int
-  #                 ClassName
-  #                 RoleName
-  #         Ref
-  #             ScalarRef[`a]
-  #             ArrayRef[`a]
-  #             HashRef[`a]
-  #             CodeRef
-  #             RegexpRef
-  #             GlobRef
-  #                 FileHandle
-  #             Object
-
-  # So the test for 
-
-  my %simple_types;
-
-  BEGIN
-  {
-    %simple_types = map { $_ => 1 }
-      qw ( Any Item Bool Undef Defined Value Str Num Int __ANON__ );
-  }
-
-########################################
-# Usage:
-#   _check_ref_attribs($class, $ref_attribs);
-# Checks that "reference" attributes from the descriptor (e.g., attr => [N])
-# are ArrayRef[]'s of simple attributes in the Moose object
-# (e.g., isa => ArrayRef['Str'])
-# Throws an exception if check fails
-#
-# where:
-#   $class is the current Moose class
-#   $ref_attribs an hashref of Moose attributes which are "ref
-#   attributes", e.g., " has 'hobbies' (isa => 'ArrayRef[Str]'); "
-#
-########################################
-sub _check_ref_attribs {
-  my ($class, $ref_attribs) = @_;
-
-  my $meta = $class->meta
-    or croak "No meta for class '$class'?";
-
-  foreach my $attrib ( keys %{ $ref_attribs } ) {
-    my $msg = "Moose class '$class' ref attrib '$attrib'";
-
-    my $constraint = $meta->find_attribute_by_name($attrib)->type_constraint
-      or croak "$msg has no type constraint";
-
-    #print "_check_ref_attribs(): $attrib $constraint\n";
-
-    if ($constraint =~ /^ArrayRef\[([^]]*)\]/ ) {
-
-      croak "$msg has bad type '$constraint' ('$1' is not a simple type)"
-        unless $simple_types{$1};
-
-      return;
-    }
-    croak "$msg must be an ArrayRef[`a] and not a '$constraint'";
-  }
-}
-
-
-########################################
-# Usage:
-#   _check_non_ref_attribs($class, $non_ref_attribs);
-# Checks that non-ref attributes from the descriptor (e.g., attr => N)
-# are indeed simple attributes in the Moose object (e.g., isa => 'Str')
-# Throws an exception if check fails
-#
-#
-# where:
-#   $class is the current Moose class
-#   $non_ref_attribs an hashref of Moose attributes which are 
-#   non-reference, or "simple" attributes like 'Str', 'Int', etc.
-#   The key is the attribute name, the value the type
-#
-########################################
-sub _check_non_ref_attribs {
-  my ($class, $attribs) = @_;
-
-  my $meta = $class->meta
-    or croak "No meta for class '$class'?";
-
-  foreach my $attrib ( keys %{ $attribs } ) {
-    my $msg = "Moose class '$class', attrib '$attrib'";
-
-    my $constraint = $meta->find_attribute_by_name($attrib)->type_constraint
-      or croak "$msg has no type (isa => ...)";
-
-    #print "_check_non_ref_attribs(): $attrib '$constraint'\n";
-
-    # kludge for Maybe[`]
-    $constraint =~ /^Maybe\[([^]]+)\]/;
-    $constraint = $1 if $1;
-
-    #print " after: $attrib '$constraint'\n";
-
-    next if $simple_types{$constraint};
-
-    $msg = "$msg has type '$constraint', but your descriptor had '$attrib => " 
-         . $attribs->{$attrib} . "'.";
-
-    $msg .= " (Did you forget the '[]' brackets?)"
-      if $constraint =~ /^ArrayRef/;
-      
-    croak $msg;
-  }
-}
-      
-} # end of local block
-
-
 1;
 
 __END__
@@ -560,14 +347,14 @@ __END__
 
 Array::To::Moose - Build Moose objects from a data array
 
-=head1 VERSION
+=head1 VERSION 
 
-This document describes Array::To::Moose version 0.0.9
+This document describes Array::To::Moose version 0.0.1
 
 =head1 SYNOPSIS
 
   use Array::To::Moose;
-  # or
+  # or 
   use Array::To::Moose qw(array_to_moose set_class_ind set_key_ind
                           throw_nonunique_keys throw_multiple_rows   );
 
@@ -579,11 +366,10 @@ C<throw_nonunique_keys()> and C<throw_multiple_rows()> if requested.
 
 C<array_to_moose()> builds Moose objects from suitably-sorted
 2-dimensional arrays of data of the type returned by, e.g.,
-L<DBI::selectall_arrayref()|DBI/selectall_arrayref>
-i.e.  a reference to an array containing
+L<DBI::selectall_arrayref()>, i.e.  a reference to an array containing
 references to an array for each row of data fetched.
 
-=head2 Example 1a
+=head2 Example 1
 
   package Car;
   use Moose;
@@ -632,8 +418,6 @@ references to an array for each row of data fetched.
 
   print $CarOwners->[2]->Cars->[1]->model; # prints "Camry"
 
-=head2 Example 1b - Hash(ref) Sub-objects
-
 In the above example, C<array_to_moose()> returns a reference to an
 B<array> of C<CarOwner> objects, C<$CarOwners>.
 
@@ -671,8 +455,8 @@ Similarly, to construct the C<Cars> sub-objects as I<hash> sub-objects
   has 'first' => (is => 'ro', isa => 'Str'         );
   has 'Cars'  => (is => 'ro', isa => 'HashRef[Car]'); # Was 'ArrayRef[Car]'
 
-and noting that the car C<make> is unique for each person in the C<$data> dataset, we
-construct the reference to an array of objects with the call:
+and noting that the car C<make> is unique within the C<$data> dataset, we
+could construct the reference to an array of objects with the call:
 
   $CarOwners = array_to_moose(
                       data => $data,
@@ -691,43 +475,6 @@ construct the reference to an array of objects with the call:
 
   print $CarOwners->[2]->Cars->{BMW}->model; # prints 'X5'
 
-=head2 Example 1c - "Simple" Reference Attributes
-
-If, instead of the car owner object containing an ArrayRef or HashRef of
-C<Car> sub-objects, it contains, say, a ArrayRef of strings representing the
-names of the car makers:
-
-  package SimpleCarOwner;
-  use Moose;
-
-  has 'last'      => (is => 'ro', isa => 'Str'          );
-  has 'first'     => (is => 'ro', isa => 'Str'          );
-  has 'CarMakers' => (is => 'ro', isa => 'ArrayRef[Str]');
-
-Using the same dataset from Example 1a, we construct an arrayref
-C<SimpleCarOwner> objects as:
-
-  $SimpleCarOwners = array_to_moose(
-                        data => $data,
-                        desc => {
-                          class     => 'SimpleCarOwner',
-                          last      => 0,
-                          first     => 1,
-                          CarMakers => [2],  # Note the '[...]' brackets
-                        }
-  );
-
-  print $SimpleCarOwners->[2]->[1];   # prints 'Toyota'
-
-I.e., when the object attribute is an I<ArrayRef> of one of the Moose "simple" types,
-e.g. C<'Str'>, C<'Num'>, C<'Bool'>,
-etc (See L<Moose::Manual::Types|THE TYPES>), then the column number should
-appear in square brackets ('C<CarMakers =E<gt> [2]>' above) to differentiate them from the bare
-types (C<last =E<gt> 0,> and C<first =E<gt> 1,> above).
-
-Note that Array::To::Moose doesn't (yet) handle the case of hashrefs of
-"simple" types, e.g., C<( isa =E<gt> "HashRef[Str]" )>
-
 =head2 Example 2 - Use with DBI
 
 The main rationale for writing C<Array::To::Moose> is to make it easy to build
@@ -743,7 +490,7 @@ I<Visit> table, which in turn would have a one-to-many relationship with
 the I<Test> table
 
 The corresponding Moose model has nested Moose objects which reflects those
-one-to-many relationships, i.e.,
+one-to-many relationships, i.e., 
 multiple Visit objects per Patient object and multiple Test objects
 per Visit object, declared as:
 
@@ -821,7 +568,7 @@ In the main program:
 
   print $patients->[2]->Visits->[0]->Tests->{BP}->result; # prints '120/80'
 
-Note: We used the Test C<name> as the key for the Visit 'C<Tests>', as the
+Note: We used the Test C<name> as the key for the Visis 'C<Tests>', as the
 tests have unique names within any one Visit.
 (See t/5.t)
 
@@ -831,21 +578,18 @@ As shown in the above examples, the general usage is:
 
   package MyClass;
   use Moose;
-  (define Moose object(s))
   ...
   use Array::To::Moose;
   ...
-  my $data_ref = selectall_arrayref($sql); # for example
+  my $data_ref = fetchall_arrayref($sql); # for example
 
   my $object_ref =  array_to_moose(
                         data => $data_ref
                         desc => {
                           class    => 'MyClass',
-                          key      => K,   # only for HashRefs
-                          attrib_1 => N1,
-                          attrib_2 => N2,
-                          ...
-                          attrib_m => [ M ],
+                          key      => <key_col>,   # only for HashRefs
+                          attrib_1 => <column_number_1>,
+                          attrib_2 => <column_number_2>,
                           ...
                           SubObject => {
                             class => 'MySubClass',
@@ -856,35 +600,28 @@ As shown in the above examples, the general usage is:
 
 Where:
 
-C<array_to_moose()> returns an array- or hash reference of C<MyClass>
-Moose objects.
-All Moose classes (C<MyClass>, C<MySubClass>, etc) must
-already have been defined by the user.
+C<$object_ref> will contain a reference to an array or hash of C<MyClass>
+Moose objects. All Moose classes (C<MyClass>, C<MySubClass>, etc) must
+already have been defined.
 
 C<$data_ref> is a reference to an array containing references to arrays of
-scalars of the kind returned by, e.g.,
+scalars (AoA) of the kind returned by
 L<DBI::selectall_arrayref()|DBI/selectall_arrayref>
 
 C<desc> (descriptor) is a reference to a hash which contains several types
 of data:
 
-C<class =E<gt>> 'MyObj' is I<required> and defines the Moose class or
+C<class =E<gt>> ... is I<required> and defines the Moose class or
 package which will contain the data. The user should have defined this class
 already.
 
-C<key =E<gt> N > is required
-if the Moose object being constructed is to be a hashref, either at
+C<key =E<gt> >... is required 
+if the Moose object being constructed is to be a hashref, either as
 the top-level Moose object returned from C<array_to_moose()> or as a
 "C<isa =E<gt> 'HashRef[...]'>" sub-object.
 
 C<attrib =E<gt> N > where C<attrib> is the name of a Moose attribute
-("C<has 'attrib' =E<gt>> ...") 
-
-C<attrib =E<gt> [ N ] > where C<attrib> is the name of a Moose "simple" sub-attribute
-("C<has =E<gt> 'attrib' ( isa =E<gt> 'ArrayRef[Type]' ...)> "), where C<Type>
-is a "simple" Moose type, e.g., C<'Str', 'Int'>, etc.
-
-In the above cases, C<N> is a positive integer containing the
+("C<has 'attrib' =E<gt>> ..."), C<N> is a positive integer containing the
 the corresponding zero-indexed
 column number in the data array where that attribute's data is to be found.
 
@@ -894,15 +631,15 @@ C<array_to_moose()> can handle three types of Moose sub-objects, i.e.:
 
 an array of sub-objects:
 
-  has => 'Sub_Obj' ( isa => 'ArrayRef[MyObj]' );
+  has => Sub_Obj ( isa => 'ArrayRef[MyObj]' );
 
 a hash of sub-objects:
 
-  has => 'Sub_Obj' ( isa => 'HashRef[MyObj]'  );
+  has => Sub_Obj ( isa => 'HashRef[MyObj]'  );
 
 or a single sub-object:
 
-  has => 'Sub_Obj' ( isa => 'MyObj'           );
+  has => Sub_Obj ( isa => 'MyObj'           );
 
 the descriptor entry for C<Sub_Obj> in each of these cases is (almost) the same:
 
@@ -921,13 +658,6 @@ the descriptor entry for C<Sub_Obj> in each of these cases is (almost) the same:
 (A C<HashRef[']> sub-object will also I<require> a
 C<key =E<gt> N> entry in the descriptor).
 
-In addition, C<array_to_moose()> can also handle C<ArrayRef>s of "simple"
-types:
-
-  has => 'Sub_Obj' ( isa => 'ArrayRef[Type]' );
-
-where C<Type> is a "simple" Moose type, e.g., C<'Str', 'Int, 'Bool'>, etc.
-
 =head2 Ordering the data
 
 C<array_to_moose()> does not sort the input data array, and does all
@@ -953,7 +683,7 @@ I<Patient> data the slowest:
 In SQL this would be accomplished by a C<SORT BY> clause, e.g.:
 
   SORT BY Patient.Key, Visit.Key, Test.Key
-
+    
 =head2 throw_nonunique_keys ()
 
 By default, C<array_to_moose()> does not check the uniqueness of hash key
@@ -1001,13 +731,13 @@ C<array_to_moose>
 will silently overwrite the 2005 Employer object with the data for the
 2006 Employer object:
 
-  print $obj->[0]->Employers->{'Acme Corp'}->year, "\n"; # prints '2006'
+  print $object->[0]->Employers->{'Acme Corp'}->year, "\n"; # prints '2006'
 
 Calling C<throw_uniq_keys()> (either with no argument, or with a non-zero
 argument) enables reporting of non-unique keys. In the above example,
 C<array_to_moose()> would exit with warning:
 
- Non-unique key 'Acme Corp' in 'Employer' class ...
+ Non-unique key 'Acme Corp' in 'Employer' class ... 
 
 Calling C<throw_uniq_keys(0)>, i.e. with an argument of zero will disable
 subsequent reporting of non-unique keys.
@@ -1102,7 +832,7 @@ For example:
                         }
   );
 
-
+ 
 =head2 Read-only Attributes
 
 One of the recommendations of L<Moose::Manual::BestPractices>
@@ -1119,12 +849,9 @@ object evaluations are carried out depth-first.
 
 =head2 Treatment of C<NULL>s
 
-C<array_to_moose()> uses
-L<Array::GroupBy::igroup_by|Array::GroupBy.pm/DESCRIPTION>
-to compare the rows in
+C<array_to_moose()> uses C<Array::GroupBy::igroup_by> to compare the rows in
 the data given in C<data =E<gt> ...>, using function
-L<Array::GroupBy::str_row_equal()|Array::GroupBy.pm/Routines_str_row_equal()_and_num_row_equal()>
-which compares the data as I<strings>.
+C<Array::GroupBy::str_row_equal()> which compares the data as I<strings>.
 
 If the data contains C<undef> values, typically returned from
 database SQL queries in which L<DBI> maps NULL values to C<undef>, when
@@ -1143,31 +870,49 @@ This truth table demonstrates the various combinations:
 
 =head1 EXPORT
 
-C<array_to_moose> by default; C<throw_nonunique_keys>, C<throw_multiple_rows>,
-C<set_class_ind> and C<set_key_ind> if requested.
+C<array_to_moose> by default; and C<set_class_ind> and
+C<set_key_ind> if requested.
 
 =head1 DIAGNOSTICS
 
-Errors in the call of C<array-to-moose()> will be caught by
-L<Params::Validate::Array>, q.v.
+set_key_ind() argument not defined
+croak "set_class_ind() argument not defined"
+croak "No class => ..."
+Class '$class' not defined
+Attribute '$name' not in '$class' object
 
-<array-to-moose> does a lot of error checking, and is probably annoyingly
-chatty. Most of the errors generated are, of course, self-explanatory :-)
+(test if 'class' or 'key' isn't also an attribute)
+the '$class' object has an attribute called '$name'
+no data in data => ...
+found a ref in attribs:
+(column#) $msg must be a +ve integer
+(column#) greater than # cols in the data
+'array_to_moose(desc => ...)' arg has an odd number of members
+empty descriptor
+no '$CLASS' defined in descriptor:
+can't delete '$CLASS' element of descriptor
+can't delete '$KEY' element of descriptor
+attribute '$name' can't be a '$ref' reference
+no attributes with column numbers in descriptor:
+Moose attribute '$attr_name' has no type
+
+
+desc generated a '", ref $sub_obj, "' object and "
+              . "not the expected array. desc = "
+
+Note that its the user's responsibility to make sure that the types of data in
+the AoA matches the Moose attributes.
 
 =head1 DEPENDENCIES
 
   Carp
-  Params::Validate::Array
+  Params::Validate
   Array::GroupBy
 
 =head1 SEE ALSO
 
 L<DBI>, L<Moose>, L<Array::GroupBy>
 
-=head1 BUGS
-
-The handling of Moose type constraints is primitive.
-
 =head1 AUTHOR
 
 Sam Brain <samb@stanford.edu>
@@ -1186,7 +931,8 @@ at your option, any later version of Perl 5 you may have available.
 
 # TODO
 #
-# test for non-square data array?
+# - catch data mismatch, eg data => contains non-numeric, but Moose attrib is
+# numeric
 #
 # - allow argument "compare => sub {...}" in array_to_moose() call to
 # allow a user-defined row-comparison routine to be passed to
@@ -1197,21 +943,19 @@ at your option, any later version of Perl 5 you may have available.
 ##### SUBROUTINE INDEX #####
 #                          #
 #   gen by index_subs.pl   #
-#   on 24 Apr 2014 21:11   #
+#   on  1 Jan 2013 21:27   #
 #                          #
 ############################
 
 
 ####### Packages ###########
 
-# Array::To::Moose ......................... 1
-#   array_to_moose ......................... 2
-#   set_class_ind .......................... 2
-#   set_key_ind ............................ 2
-#   throw_multiple_rows .................... 2
-#   throw_nonunique_keys ................... 2
-#   _check_descriptor ...................... 4
-#   _check_non_ref_attribs ................. 9
-#   _check_ref_attribs ..................... 8
-#   _check_subobj .......................... 6
+# Array::To::Moose ....................... 1
+#   array_to_moose ....................... 3
+#   set_class_ind ........................ 2
+#   set_key_ind .......................... 1
+#   throw_multiple_rows .................. 2
+#   throw_nonunique_keys ................. 2
+#   _check_attribs ....................... 2
+#   _check_types ......................... 5
 
